/**
 * Copyright (c) 2013-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 */

'use strict'; // eslint-disable-line strict

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

module.exports = function (context) {
  var t = context.types;

  var TransformedTernary = function () {
    function TransformedTernary(expression, generateUid) {
      _classCallCheck(this, TransformedTernary);

      this.generateUid = generateUid;

      this.uids = [];
      this.expression = expression;
      this.deepestTernary = null;
      this.deepestExpression = expression;
    }

    _createClass(TransformedTernary, [{
      key: 'constructTernary',
      value: function constructTernary(oldExpression, newExpression, uid) {
        return t.ConditionalExpression(t.BinaryExpression('!=', t.AssignmentExpression('=', t.Identifier(uid), oldExpression), t.NullLiteral()), newExpression, t.Identifier(uid));
      }
    }, {
      key: 'addLevel',
      value: function addLevel(expression, uid) {
        var ternary = this.constructTernary(this.deepestExpression, expression, uid);
        if (this.deepestTernary === null) {
          this.expression = ternary;
        } else {
          this.deepestTernary.consequent = ternary;
        }
        this.deepestTernary = ternary;
        this.deepestExpression = expression;
      }
    }, {
      key: 'appendMethodCall',
      value: function appendMethodCall(args) {
        var uid = this.generateUid('ref').name;
        var callExpression = t.CallExpression(t.Identifier(uid), args || []);
        this.addLevel(callExpression, uid);
        this.uids.push(uid);
      }
    }, {
      key: 'appendPropertyAccess',
      value: function appendPropertyAccess(property, computed) {
        var uid = this.generateUid('ref').name;
        var accessedProperty = t.MemberExpression(t.Identifier(uid), property, computed);
        this.addLevel(accessedProperty, uid);
        this.uids.push(uid);
      }
    }]);

    return TransformedTernary;
  }();

  function checkIdxArguments(args) {
    if (args.length !== 2) {
      throw new Error('The `idx` function takes exactly two arguments.');
    }
    var arrowFunction = args[1];
    if (!t.isArrowFunctionExpression(arrowFunction)) {
      throw new Error('The second argument supplied to `idx` must be an arrow function.');
    }
    if (!t.isExpression(arrowFunction.body)) {
      throw new Error('The body of the arrow function supplied to `idx` must be a single ' + 'expression (without curly braces).');
    }
    if (arrowFunction.params.length !== 1) {
      throw new Error('The arrow function supplied to `idx` must take exactly one parameter.');
    }
    var bodyChainBase = getExpressionChainBase(arrowFunction.body);
    if (!t.isIdentifier(arrowFunction.params[0]) || !t.isIdentifier(bodyChainBase) || arrowFunction.params[0].name !== bodyChainBase.name) {
      throw new Error('The parameter of the arrow function supplied to `idx` must match ' + 'the base of the body expression.');
    }
  }

  function getExpressionChainBase(node) {
    if (t.isCallExpression(node)) {
      return getExpressionChainBase(node.callee);
    } else if (t.isMemberExpression(node)) {
      return getExpressionChainBase(node.object);
    } else {
      return node;
    }
  }

  function isIdxCall(node) {
    return t.isCallExpression(node) && t.isIdentifier(node.callee) && node.callee.name === 'idx';
  }

  function constructTernary(base, node, generateUid) {
    if (t.isCallExpression(node)) {
      var transformedObject = constructTernary(base, node.callee, generateUid);
      transformedObject.appendMethodCall(node.arguments);
      return transformedObject;
    } else if (t.isMemberExpression(node)) {
      var _transformedObject = constructTernary(base, node.object, generateUid);
      _transformedObject.appendPropertyAccess(node.property, node.computed);
      return _transformedObject;
    } else {
      return new TransformedTernary(base, generateUid);
    }
  }

  var idxVisitor = {
    CallExpression: function CallExpression(path) {
      var node = path.node;
      if (isIdxCall(node)) {
        checkIdxArguments(node.arguments);
        var ternary = constructTernary(node.arguments[0], node.arguments[1].body, path.scope.generateUidIdentifier.bind(path.scope));
        path.replaceWith(ternary.expression);
        for (var ii = 0; ii < ternary.uids.length; ii++) {
          var uid = ternary.uids[ii];
          path.scope.push({ id: t.Identifier(uid) });
        }
      }
    }
  };

  return {
    visitor: {
      Program: function Program(path) {
        // We're very strict about the shape of idx. Some transforms, like
        // "babel-plugin-transform-async-to-generator", will convert arrow
        // functions inside async functions into regular functions. So we do
        // our transformation before any one else interferes.
        path.traverse(idxVisitor);
      }
    }
  };
};